@@ -40,8 +40,6 @@ struct _MetaWorkspace
   MetaScreen *screen;
   
   GList *windows;
-
-  char *name;
   
   MetaRectangle work_area;
   guint work_area_invalid : 1;
@@ -60,7 +58,6 @@ gboolean       meta_workspace_contains_window (MetaWorkspace *workspace,
                                                MetaWindow  *window);
 void           meta_workspace_activate      (MetaWorkspace *workspace);
 int            meta_workspace_index         (MetaWorkspace *workspace);
-int            meta_workspace_screen_index  (MetaWorkspace *workspace);
 GList*         meta_workspace_list_windows  (MetaWorkspace *workspace);
 
 void meta_workspace_invalidate_work_area (MetaWorkspace *workspace);
@@ -71,8 +68,7 @@ void meta_workspace_get_work_area        (MetaWorkspace *workspace,
 MetaWorkspace* meta_workspace_get_neighbor (MetaWorkspace      *workspace,
                                             MetaMotionDirection direction);
 
-void meta_workspace_set_name (MetaWorkspace *workspace,
-                              const char    *name);
+const char* meta_workspace_get_name (MetaWorkspace *workspace);
 
 #endif
 
